==> exrm-0.17.1.tar: lib/exrm/plugin.ex <==
defmodule ReleaseManager.Plugin do
  @moduledoc """
  This module provide a simple way to add additional steps to
  the release task.

  You can define your own plugins using the sample definition below. Note that
  the module namespace must be nested under `ReleaseManager.Plugin.*`.

      defmodule ReleaseManager.Plugin.Hello do
        use ReleaseManager.Plugin

        def before_release(%Config{} = config) do
          info "This is executed just prior to compiling the release"
        end

        def after_release(%Config{} = config) do
          info "This is executed just after compiling the release"
        end

        def after_package(%Config{} = config) do
          info "This is executed just after packaging the release"
        end

        def after_cleanup(_args) do
          info "This is executed just after running cleanup"
        end
      end

  A couple things are imported or aliased for you. Those things are:

    - The ReleaseManager.Config struct is aliased for you to just Config
    - `debug/1`, `info/1`, `warn/1`, `notice/1`, and `error/1` are imported for you.
      These should be used to do any output for the user.

  `before_release/1` and `after_release/1` are required callbacks, and will each be passed a
  `Config` struct, containing the configuration for the release task. You can choose
  to return the config struct modified or unmodified, or not at all. In the former case,
  any modifications you made will be passed on to the remaining plugins and the final
  release task. The required callback `after_cleanup/1` is passed the command line arguments.
  The return value is not used.

  All plugins are executed just prior, and just after compiling the release, as the name of
  the callbacks reflect. The `before_release/1` callback is called after some internal tasks,
  such as generating the sys.config and others.
  """
  use Behaviour

  @doc """
  A plugin needs to implement `before_release/1`, and `after_release/1`
  both of which receive a %ReleaseManager.Config struct, as well as `after_cleanup/1`, which
  receives the arguments given for the command as a list of strings.
  """
  defcallback before_release(ReleaseManager.Config.t) :: any
  defcallback after_release(ReleaseManager.Config.t) :: any
  defcallback after_package(ReleaseManager.Config.t) :: any
  defcallback after_cleanup([String.t]) :: any

  @doc false
  defmacro __using__(_opts) do
    quote do
      @behaviour ReleaseManager.Plugin
      alias  ReleaseManager.Config
      import ReleaseManager.Utils, only: [debug: 1, info: 1, warn: 1, notice: 1, error: 1]

      Module.register_attribute __MODULE__, :name, accumulate: false, persist: true
      Module.register_attribute __MODULE__, :moduledoc, accumulate: false, persist: true
      Module.register_attribute __MODULE__, :shortdoc, accumulate: false, persist: true
    end
  end

  @doc """
  Loads all plugins in all code paths.
  """
  @spec load_all() :: [] | [atom]
  def load_all, do: load_plugins(:code.get_path)

  @doc """
  Loads all plugins in the given `paths`.
  """
  @spec load_plugins([binary]) :: [] | [atom]
  def load_plugins(paths) do
    Enum.reduce(paths, [], fn(path, matches) ->
      {:ok, files} = :erl_prim_loader.list_dir(path)
      Enum.reduce(files, matches, &match_plugins/2)
    end)
  end

  @re_pattern Regex.re_pattern(~r/Elixir\.ReleaseManager\.Plugin\..*\.beam$/)

  @spec match_plugins(char_list, [atom]) :: [atom]
  defp match_plugins(filename, modules) do
    if :re.run(filename, @re_pattern, [capture: :none]) == :match do
      mod = :filename.rootname(filename, '.beam') |> List.to_atom
      if Code.ensure_loaded?(mod), do: [mod | modules], else: modules
    else
      modules
    end
  end
end

==> exrm-0.18.2.tar: lib/exrm/plugin.ex <==
defmodule ReleaseManager.Plugin do
  @moduledoc """
  This module provide a simple way to add additional steps to
  the release task.

  You can define your own plugins using the sample definition below. Note that
  the module namespace must be nested under `ReleaseManager.Plugin.*`.

      defmodule ReleaseManager.Plugin.Hello do
        use ReleaseManager.Plugin

        def before_release(%Config{} = config) do
          info "This is executed just prior to compiling the release"
        end

        def after_release(%Config{} = config) do
          info "This is executed just after compiling the release"
        end

        def after_package(%Config{} = config) do
          info "This is executed just after packaging the release"
        end

        def after_cleanup(_args) do
          info "This is executed just after running cleanup"
        end
      end

  A couple things are imported or aliased for you. Those things are:

    - The ReleaseManager.Config struct is aliased for you to just Config
    - `debug/1`, `info/1`, `warn/1`, `notice/1`, and `error/1` are imported for you.
      These should be used to do any output for the user.

  `before_release/1` and `after_release/1` are required callbacks, and will each be passed a
  `Config` struct, containing the configuration for the release task. You can choose
  to return the config struct modified or unmodified, or not at all. In the former case,
  any modifications you made will be passed on to the remaining plugins and the final
  release task. The required callback `after_cleanup/1` is passed the command line arguments.
  The return value is not used.

  All plugins are executed just prior, and just after compiling the release, as the name of
  the callbacks reflect. The `before_release/1` callback is called after some internal tasks,
  such as generating the sys.config and others.
  """
  use Behaviour

  @doc """
  A plugin needs to implement `before_release/1`, and `after_release/1`
  both of which receive a %ReleaseManager.Config struct, as well as `after_cleanup/1`, which
  receives the arguments given for the command as a list of strings.
  """
  defcallback before_release(ReleaseManager.Config.t) :: any
  defcallback after_release(ReleaseManager.Config.t) :: any
  defcallback after_package(ReleaseManager.Config.t) :: any
  defcallback after_cleanup([String.t]) :: any

  @doc false
  defmacro __using__(_opts) do
    quote do
      @behaviour ReleaseManager.Plugin
      alias  ReleaseManager.Config
      import ReleaseManager.Utils, only: [debug: 1, info: 1, warn: 1, notice: 1, error: 1]

      Module.register_attribute __MODULE__, :name, accumulate: false, persist: true
      Module.register_attribute __MODULE__, :moduledoc, accumulate: false, persist: true
      Module.register_attribute __MODULE__, :shortdoc, accumulate: false, persist: true
    end
  end

  @doc """
  Loads all plugins in all code paths.
  """
  @spec load_all() :: [] | [atom]
  def load_all, do: get_plugins(ReleaseManager.Plugin)

  # Loads all modules that extend a given module in the current code path.
  #
  # The convention is that it will fetch modules with the same root namespace,
  # and that are suffixed with the name of the module they are extending.
  @spec get_plugins(atom) :: [] | [atom]
  defp get_plugins(plugin_type) when is_atom(plugin_type) do
    available_modules(plugin_type) |> Enum.reduce([], &load_plugin/2)
  end

  defp load_plugin(module, modules) do
    if Code.ensure_loaded?(module), do: [module | modules], else: modules
  end

  defp available_modules(plugin_type) do
    apps_path = Mix.Project.build_path |> Path.join("lib")
    apps      = case apps_path |> File.ls do
      {:ok, apps} -> apps
      {:error, _} -> []
    end
    apps
    |> Enum.map(&(Path.join([apps_path, &1, "ebin"])))
    |> Enum.map(fn app_path -> app_path |> File.ls! |> Enum.map(&(Path.join(app_path, &1))) end)
    |> Enum.flat_map(&(&1))
    |> Enum.filter(&(String.ends_with?(&1, ".beam")))
    |> Enum.map(fn path ->
      {:ok, {module, chunks}} = :beam_lib.chunks('#{path}', [:attributes])
      {module, get_in(chunks, [:attributes, :behaviour])}
    end)
    |> Enum.filter(fn {_module, behaviours} ->
      is_list(behaviours) && plugin_type in behaviours
    end)
    |> Enum.map(fn {module, _} -> module end)
  end
end
